import re
from datetime import datetime
from decimal import Decimal as D

from django.utils.dateparse import parse_datetime

from gisserver.exceptions import ExternalParsingError

RE_FLOAT = re.compile(r"\A[0-9]+(\.[0-9]+)\Z")


def auto_cast(value: str):
    """Automatically cast a value to a scalar."""
    if value.isdigit():
        return int(value)
    elif RE_FLOAT.match(value):
        return D(value)
    elif "T" in value:
        try:
            return parse_iso_datetime(value)
        except ValueError:
            pass

    return value


def parse_iso_datetime(raw_value: str) -> datetime:
    value = parse_datetime(raw_value)
    if value is None:
        raise ExternalParsingError(
            "Date must be in YYYY-MM-DD HH:MM[:ss[.uuuuuu]][TZ] format."
        )
    return value


def parse_bool(raw_value: str):
    if raw_value in ("true", "1"):
        return True
    elif raw_value in ("false", "0"):
        return False
    else:
        raise ExternalParsingError(f"Can't cast '{raw_value}' to boolean")



"""The base protocol to implement an operation.

All operations extend from an WFSMethod class.
This defines the parameters and output formats of the method.
This introspection data is also parsed by the GetCapabilities call.
"""
from __future__ import annotations

import math
import re
from dataclasses import dataclass, field
from typing import Any, Callable

from django.core.exceptions import ImproperlyConfigured, SuspiciousOperation
from django.http import HttpResponse
from django.template.loader import render_to_string

from gisserver.exceptions import (
    ExternalParsingError,
    InvalidParameterValue,
    MissingParameterValue,
    OperationParsingFailed,
)
from gisserver.features import FeatureType

NoneType = type(None)

RE_SAFE_FILENAME = re.compile(r"\A[A-Za-z0-9]+[A-Za-z0-9.]*")  # no dot at the start.


@dataclass(frozen=True)
class Parameter:
    """The definition of an parameter for an WFS method.

    These parameters should be defined in the
    WFSMethod parameters attribute / get_parameters().

    These fields is used to parse the incoming request.
    The GetCapabilities method also reads the metadata of this value.
    """

    #: Name of the parameter (using the XML casing style)
    name: str

    #: Alias name (e.g. typenames/typename)
    alias: str | None = None

    #: Whether the parameter is required
    required: bool = False

    #: Optional parser callback to convert the parameter into a Python type
    parser: Callable[[str], Any] = None

    #: Whether to list this parameter in GetCapabilities
    in_capabilities: bool = False

    #: List of allowed values (also shown in GetCapabilities)
    allowed_values: list | tuple | set | NoneType = None

    #: Default value if it's not given
    default: Any = None

    #: Overridable dict for error messages
    error_messages: dict = field(default_factory=dict)

    def value_from_query(self, KVP: dict):  # noqa: C901
        """Parse a request variable using the type definition.

        This uses the dataclass settings to parse the incoming request value.
        """
        # URL-based key-value-pair parameters use uppercase.
        kvp_name = self.name.upper()
        value = KVP.get(kvp_name)
        if not value and self.alias:
            value = KVP.get(self.alias.upper())

        # Check required field settings, both empty and missing value are treated the same.
        if not value:
            if not self.required:
                return self.default
            elif value is None:
                raise MissingParameterValue(self.name, f"Missing {kvp_name} parameter")
            else:
                raise InvalidParameterValue(self.name, f"Empty {kvp_name} parameter")

        # Allow conversion into a python object
        if self.parser is not None:
            try:
                value = self.parser(value)
            except ExternalParsingError as e:
                raise OperationParsingFailed(
                    self.name, f"Unable to parse {kvp_name} argument: {e}"
                ) from None
            except (TypeError, ValueError, NotImplementedError) as e:
                # TypeError/ValueError are raised by most handlers for unexpected data
                # The NotImplementedError can be raised by fes parsing.
                raise InvalidParameterValue(
                    self.name, f"Invalid {kvp_name} argument: {e}"
                ) from None

        # Validate against value choices
        self.validate_value(value)

        return value

    def validate_value(self, value):
        """Validate the parsed value.
        This method can be overwritten by a subclass if needed.
        """
        if self.allowed_values is not None and value not in self.allowed_values:
            msg = self.error_messages.get(
                "invalid", "Invalid value for {name}: {value}"
            )
            raise InvalidParameterValue(
                self.name, msg.format(name=self.name, value=value)
            )


class UnsupportedParameter(Parameter):
    def value_from_query(self, KVP: dict):
        kvp_name = self.name.upper()
        if kvp_name in KVP:
            raise InvalidParameterValue(
                self.name, f"Support for {self.name} is not implemented!"
            )
        return None


class OutputFormat:
    """Declare an output format for the method.

    These formats should be used in the ``output_formats`` section of the WFSMethod.
    """

    def __init__(
        self,
        content_type,
        renderer_class=None,
        max_page_size=None,
        title=None,
        **extra,
    ):
        """
        :param content_type: The MIME-type used in the request to select this type.
        :param renderer_class: The class that performs the output rendering.
            If it's not given, the operation renders it's output using an XML template.
        :param max_page_size: Used to override the ``max_page_size`` of the renderer_class.
        :param title: A human-friendly name for a HTML overview page.
        :param extra: Any additional key-value pairs for the definition.
            Could include ``subtype`` as a shorter alias for the MIME-type.
        """
        self.content_type = content_type
        self.extra = extra
        self.subtype = self.extra.get("subtype")
        self.renderer_class = renderer_class
        self.title = title
        self._max_page_size = max_page_size

    def matches(self, value):
        """Test whether the 'value' is matched by this object."""
        return self.content_type == value or self.subtype == value

    @property
    def identifier(self):
        """The identifier to use in templates as OUTPUTFORMAT input value."""
        return self.subtype or self.content_type

    @property
    def max_page_size(self):
        """Override the default max page size"""
        if self._max_page_size:
            return self._max_page_size
        elif self.renderer_class is not None:
            return self.renderer_class.max_page_size
        else:
            return None

    @property
    def has_infinite_page_size(self):
        """Return whether the output format can be unpaginated."""
        return self.max_page_size == math.inf

    def __str__(self):
        extra = "".join(f"; {name}={value}" for name, value in self.extra.items())
        return f"{self.content_type}{extra}"

    def __repr__(self):
        return f"<OutputFormat: {self}>"


class WFSMethod:
    """Basic interface to implement an WFS method.

    Each operation in this GIS-server extends from this base class. This class
    also exposes all requires metadata for the GetCapabilities request.
    """

    do_not_call_in_templates = True  # avoid __call__ execution in Django templates

    #: List the suported parameters for this method, extended by get_parameters()
    parameters: list[Parameter] = []

    #: List the supported output formats for this method.
    output_formats: list[OutputFormat] = []

    #: Default template to use for rendering
    xml_template_name = None

    #: Default content-type for render_xml()
    xml_content_type = "text/xml; charset=utf-8"

    def __init__(self, view):
        self.view = view  # an gisserver.views.GISView
        self.namespaces = {
            # Default namespaces for the incoming request:
            "http://www.w3.org/XML/1998/namespace": "xml",
            "http://www.opengis.net/wfs/2.0": "wfs",
            "http://www.opengis.net/gml/3.2": "gml",
            self.view.xml_namespace: "app",
        }

    def get_parameters(self):
        """Dynamically return the supported parameters for this method."""
        parameters = [
            # Always add SERVICE and VERSION as required parameters
            # Part of BaseRequest:
            Parameter(
                "service",
                required=not bool(self.view.default_service),
                in_capabilities=True,
                allowed_values=list(self.view.accept_operations.keys()),
                default=self.view.default_service,  # can be None or e.g. "WFS"
                error_messages={"invalid": "Unsupported service type: {value}."},
            ),
            Parameter(
                "version",
                required=True,  # Mandatory except for GetCapabilities
                allowed_values=self.view.accept_versions,
                error_messages={
                    "invalid": "WFS Server does not support VERSION {value}.",
                },
            ),
        ] + self.parameters

        if self.output_formats:
            parameters += [
                # Part of StandardPresentationParameters:
                Parameter(
                    "outputFormat",
                    in_capabilities=True,
                    allowed_values=self.output_formats,
                    parser=self._parse_output_format,
                    default=self.output_formats[0],
                )
            ]

        return parameters

    def _parse_output_format(self, value) -> OutputFormat:
        """Select the proper OutputFormat object based on the input value"""
        value = value.replace(" ", "+")  # allow application/gml+xml on the KVP.
        try:
            return next(o for o in self.output_formats if o.matches(value))
        except StopIteration:
            raise InvalidParameterValue(
                "outputformat",
                f"'{value}' is not a permitted output format for this operation.",
            ) from None

    def _parse_namespaces(self, value) -> dict[str, str]:
        """Parse the namespaces definition.

        The NAMESPACES parameter defines which namespaces are used in the KVP request.
        When this parameter is not given, the default namespaces are assumed.
        """
        if not value:
            return {}

        # example single value: xmlns(http://example.org)
        # or: namespaces=xmlns(xml,http://www.w3.org/...),xmlns(wfs,http://www.opengis.net/...)
        tokens = value.split(",")

        namespaces = {}
        tokens = iter(tokens)
        for prefix in tokens:
            if not prefix.startswith("xmlns("):
                raise InvalidParameterValue(
                    "namespaces", f"Expected xmlns(...) format: {value}"
                )
            if prefix.endswith(")"):
                # xmlns(http://...)
                prefix = ""
                uri = prefix[6:-1]
            else:
                uri = next(tokens, "")
                if not uri.endswith(")"):
                    raise InvalidParameterValue(
                        "namespaces", f"Expected xmlns(prefix,uri) format: {value}"
                    )
                prefix = prefix[6:]
                uri = uri[:-1]

            namespaces[uri] = prefix

        return namespaces

    def parse_request(self, KVP: dict) -> dict[str, Any]:
        """Parse the parameters of the request"""
        self.namespaces.update(self._parse_namespaces(KVP.get("NAMESPACES")))
        param_values = {
            param.name: param.value_from_query(KVP) for param in self.get_parameters()
        }
        param_values["NAMESPACES"] = self.namespaces

        for param in self.get_parameters():
            param_values[param.name] = param.value_from_query(KVP)

        # Update version if requested.
        # This is stored on the view, so exceptions also use it.
        if param_values.get("version"):
            self.view.set_version(param_values["version"])

        self.validate(**param_values)
        return param_values

    def validate(self, **params):
        """Perform final request parameter validation before the method is called"""
        pass

    def __call__(self, **params):
        """Default call implementation: render an XML template."""
        context = self.get_context_data(**params)
        if "outputFormat" in params:
            output_format: OutputFormat = params["outputFormat"]

            if output_format.renderer_class is not None:
                # Streaming HTTP responses, e.g. GML32/GeoJSON output:
                renderer = output_format.renderer_class(self, **context)
                return renderer.get_response()

        return self.render_xml(context, **params)

    def get_context_data(self, **params):
        """Collect all arguments to use for rendering the XML template"""
        return {}

    def render_xml(self, context, **params):
        """Shortcut to render XML.

        This is the default method when the OutputFormat class doesn't have a renderer_class
        """
        return HttpResponse(
            render_to_string(
                self._get_xml_template_name(),
                context={
                    "view": self.view,
                    "app_xml_namespace": self.view.xml_namespace,
                    "server_url": self.view.server_url,
                    **context,
                },
            ),
            content_type=self.xml_content_type,
        )

    def _get_xml_template_name(self):
        """Generate the XML template name for this operation, and check it's file pattern"""
        service = self.view.KVP["SERVICE"].lower()
        template_name = (
            f"gisserver/{service}/{self.view.version}/{self.xml_template_name}"
        )

        # Since 'service' and 'version' are based on external input,
        # these values are double checked again to avoid remove file inclusion.
        if not RE_SAFE_FILENAME.match(service) or not RE_SAFE_FILENAME.match(
            self.view.version
        ):
            raise SuspiciousOperation(
                f"Refusing to render template name {template_name}"
            )

        return template_name


class WFSTypeNamesMethod(WFSMethod):
    """A base method that also resolved the TYPENAMES parameter."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        # Retrieve the feature types directly, as these are needed during parsing.
        # This is not delayed until _parse_type_names() as that wraps any TypeError
        # from view.get_feature_types() as an InvalidParameterValue exception.
        self.all_feature_types = self.view.get_feature_types()
        self.all_feature_types_by_name = _get_feature_types_by_name(
            self.all_feature_types
        )

    def get_parameters(self):
        return super().get_parameters() + [
            # QGis sends both TYPENAME (wfs 1.x) and TYPENAMES (wfs 2.0) for DescribeFeatureType
            # typeNames is not required when a ResourceID / GetFeatureById is specified.
            Parameter(
                "typeNames",
                alias="typeName",  # WFS 1.0 name, but still needed for CITE tests
                required=False,  # sometimes required, depends on other parameters
                parser=self._parse_type_names,
            ),
        ]

    def _parse_type_names(self, type_names) -> list[FeatureType]:
        """Find the requested feature types by name"""
        if "(" in type_names:
            # This allows to perform multiple queries in a single request:
            # TYPENAMES=(A)(B)&FILTER=(filter for A)(filter for B)
            raise OperationParsingFailed(
                "typenames",
                "Parameter lists to perform multiple queries are not supported yet.",
            )

        return [
            self._parse_type_name(name, locator="typenames")
            for name in type_names.split(",")
        ]

    def _parse_type_name(self, name, locator="typename") -> FeatureType:
        """Find the requested feature type for a type name"""
        app_prefix = self.namespaces[self.view.xml_namespace]
        if name.startswith(f"{app_prefix}:"):
            local_name = name[len(app_prefix) + 1 :]  # strip our XML prefix
        else:
            local_name = name

        try:
            return self.all_feature_types_by_name[local_name]
        except KeyError:
            raise InvalidParameterValue(
                locator,
                f"Typename '{name}' doesn't exist in this server. "
                f"Please check the capabilities and reformulate your request.",
            ) from None


def _get_feature_types_by_name(feature_types) -> dict[str, FeatureType]:
    """Create a lookup for feature types by name."""
    features_by_name = {ft.name: ft for ft in feature_types}

    # Check against bad configuration
    if len(features_by_name) != len(feature_types):
        all_names = [ft.name for ft in feature_types]
        duplicates = ", ".join(sorted({n for n in all_names if all_names.count(n) > 1}))
        raise ImproperlyConfigured(f"FeatureType names should be unique: {duplicates}")

    return features_by_name



"""These classes map to the FES 2.0 specification for identifiers.
The class names are identical to those in the FES spec.
"""
from __future__ import annotations

from dataclasses import dataclass
from datetime import datetime
from enum import Enum

from django.db.models import Q

from gisserver import conf
from gisserver.exceptions import ExternalValueError
from gisserver.parsers.base import BaseNode, tag_registry
from gisserver.parsers.tags import expect_tag, get_attribute
from gisserver.parsers.values import auto_cast, parse_iso_datetime
from gisserver.types import FES20

NoneType = type(None)


class VersionActionTokens(Enum):
    FIRST = "FIRST"
    LAST = "LAST"
    ALL = "ALL"
    NEXT = "NEXT"
    PREVIOUS = "PREVIOUS"


class Id(BaseNode):
    """Abstract base class, as defined by FES spec."""

    xml_ns = FES20

    #: Tell which the type this ID belongs to, needs to be overwritten.
    type_name = ...  # need to be defined by subclass!

    def build_query(self, compiler) -> Q:
        raise NotImplementedError()


@dataclass
@tag_registry.register("ResourceId")
class ResourceId(Id):
    """The <fes:ResourceId> element."""

    rid: str
    version: int | datetime | VersionActionTokens | NoneType = None
    startTime: datetime | None = None
    endTime: datetime | None = None

    def __post_init__(self):
        try:
            self.type_name, self.id = self.rid.rsplit(".", 1)
        except ValueError:
            if conf.GISSERVER_WFS_STRICT_STANDARD:
                raise ExternalValueError("Expected typename.id format") from None

            # This should end in a 404 instead.
            self.type_name = None
            self.id = None

    @classmethod
    @expect_tag(FES20, "ResourceId", leaf=True)
    def from_xml(cls, element):
        version = element.get("version")
        startTime = element.get("startTime")
        endTime = element.get("endTime")

        if version:
            version = auto_cast(version)

        return cls(
            rid=get_attribute(element, "rid"),
            version=version,
            startTime=parse_iso_datetime(startTime) if startTime else None,
            endTime=parse_iso_datetime(endTime) if endTime else None,
        )

    def build_query(self, compiler=None) -> Q:
        """Render the SQL filter"""
        if self.startTime or self.endTime or self.version:
            raise NotImplementedError(
                "No support for <fes:ResourceId> startTime/endTime/version attributes"
            )

        lookup = Q(pk=self.id or self.rid)
        if compiler is not None:
            # When the
            # NOTE: type_name is currently read by the IdOperator that contains this object,
            # This code path only happens for stand-alone KVP invocation.
            compiler.add_lookups(lookup, type_name=self.type_name)
        return lookup



"""Output rendering logic for GeoJSON."""
from __future__ import annotations

import csv
from datetime import datetime

from django.conf import settings
from django.db import models
from django.utils.timezone import utc

from gisserver import conf
from gisserver.db import (
    AsEWKT,
    build_db_annotations,
    get_db_annotation,
    get_db_geometry_selects,
)
from gisserver.features import FeatureType
from gisserver.geometries import CRS
from gisserver.types import XsdComplexType, XsdElement

from .base import OutputRenderer
from .buffer import StringBuffer


class CSVRenderer(OutputRenderer):
    """Fast CSV renderer, using a stream response.

    The complex encoding bits are handled by the "csv" library.
    """

    content_type = "text/csv; charset=utf-8"
    content_disposition = 'attachment; filename="{typenames} {page} {date}.csv"'
    max_page_size = conf.GISSERVER_CSV_MAX_PAGE_SIZE

    #: The outputted CSV dialect. This can be a csv.Dialect subclass
    #: or one of the registered names like: "unix", "excel", "excel-tab"
    dialect = "unix"

    @classmethod
    def decorate_queryset(
        cls,
        feature_type: FeatureType,
        queryset: models.QuerySet,
        output_crs: CRS,
        **params,
    ) -> models.QuerySet:
        """Make sure relations are included with select-related to avoid N-queries.
        Using prefetch_related() isn't possible with .iterator().
        """
        xsd_type: XsdComplexType = feature_type.xsd_type

        # Take all relations that are expanded to complex elements,
        # and all relations that are fetched for flattened elements.
        related = {
            xsd_element.orm_path
            for xsd_element in xsd_type.complex_elements
            if not xsd_element.is_many
        } | {
            xsd_element.orm_relation[0]
            for xsd_element in xsd_type.flattened_elements
            if not xsd_element.is_many
        }
        if related:
            queryset = queryset.select_related(*related)

        return queryset

    def render_stream(self):
        output = StringBuffer()
        writer = csv.writer(output, dialect=self.dialect)

        is_first_collection = True
        for sub_collection in self.collection.results:
            if is_first_collection:
                is_first_collection = False
            else:
                # Multiple feature types requested, add newlines to separate them
                output.write("\n\n")

            # Write the header
            fields = [
                f
                for f in sub_collection.feature_type.xsd_type.elements
                if not f.is_many
            ]
            writer.writerow(self.get_header(fields))

            # By using .iterator(), the results are streamed with as little memory as
            # possible. Doing prefetch_related() is not possible now. That could only
            # be implemented with cursor pagination for large sets for 1000+ results.
            for instance in sub_collection.iterator():
                writer.writerow(self.get_row(instance, fields))

                # Only perform a 'yield' every once in a while,
                # as it goes back-and-forth for writing it to the client.
                if output.is_full():
                    yield output.flush()

        yield output.flush()

    def render_exception(self, exception: Exception):
        """Render the exception in a format that fits with the output."""
        if settings.DEBUG:
            return f"\n\n{exception.__class__.__name__}: {exception}\n"
        else:
            return f"\n\n{exception.__class__.__name__} during rendering!\n"

    def get_header(self, fields: list[XsdElement]) -> list[str]:
        """Return all field names."""
        names = []
        append = names.append
        for field in fields:
            if field.type.is_complex_type:
                # Expand complex types
                for sub_field in field.type.elements:
                    append(f"{field.name}.{sub_field.name}")
            else:
                append(field.name)

        return names

    def get_row(self, instance: models.Model, fields: list[XsdElement]):
        """Return all field values for a single row."""
        values = []
        append = values.append
        for field in fields:
            if field.is_geometry:
                append(self.render_geometry(instance, field))
                continue

            value = field.get_value(instance)
            if field.type.is_complex_type:
                for sub_field in field.type.elements:
                    append(sub_field.get_value(value))
            elif isinstance(value, list):
                # Array field
                append(",".join(map(str, value)))
            elif isinstance(value, datetime):
                append(str(value.astimezone(utc)))
            else:
                append(value)
        return values

    def render_geometry(self, instance: models.Model, field: XsdElement):
        """Render the contents of a geometry value."""
        return field.get_value(instance)


class DBCSVRenderer(CSVRenderer):
    """Further optimized CSV renderer that uses the database to render EWKT.
    This is about 40% faster then calling the GEOS C-API from python.
    """

    @classmethod
    def decorate_queryset(
        cls,
        feature_type: FeatureType,
        queryset: models.QuerySet,
        output_crs: CRS,
        **params,
    ) -> models.QuerySet:
        queryset = super().decorate_queryset(
            feature_type, queryset, output_crs, **params
        )

        # Instead of reading the binary geometry data,
        # ask the database to generate EWKT data directly.
        geo_selects = get_db_geometry_selects(
            feature_type.xsd_type.geometry_elements, output_crs
        )
        if geo_selects:
            queryset = queryset.defer(*geo_selects.keys()).annotate(
                **build_db_annotations(geo_selects, "_as_ewkt_{name}", AsEWKT)
            )

        return queryset

    def render_geometry(self, instance: models.Model, field: XsdElement):
        return get_db_annotation(instance, field.name, "_as_ewkt_{name}")



"""Internal XSD type definitions.

These types are the internal schema definition on which all output is generated.

The end-users of this library typically create a WFS feature type definition by using
the :class:~gisserver.features.FeatureType` / :class:`~gisserver.features.FeatureField` classes.

The feature type classes use the model metadata to construct the internal XMLSchema structure.
Nearly all WFS requests are handled by walking this structure (like ``DescribeFeatureType``
or ``GetFeature``). The rendered output is created by walking through this structure,
and writing the XML elements. All search queries (using XPath) are processed by
resolving the corresponding element/attributes, to find to the underlying Django model field.

The structure has the following elements:

* :class:`XsdElement` to define the schema of a single XML element (``<xsd:element>``).
* :class:`XsdAttribute` to define the schema of a single XML attribute (``<xsd:attribute>``).

Each XMLSchema node defines it's "data type" as either:

* :class:`XsdTypes` for simple well-known data types (e.g. text/int, etc..)
* :class:`XsdComplexType` for a complete class definition (holding elements and attributes)

Custom field types could also generate these field types.
"""
from __future__ import annotations

import operator
import re
from dataclasses import dataclass, field
from decimal import Decimal as D
from enum import Enum
from typing import TYPE_CHECKING

from django.conf import settings
from django.contrib.gis.db.models import F, GeometryField
from django.core.exceptions import ObjectDoesNotExist, ValidationError
from django.db import models
from django.db.models import Q
from django.db.models.fields.related import (  # Django 2 imports
    ForeignObjectRel,
    RelatedField,
)
from django.utils import dateparse
from django.utils.functional import cached_property

from gisserver.exceptions import ExternalParsingError, OperationProcessingFailed
from gisserver.geometries import CRS, WGS84  # noqa: F401 / for backwards compatibility

try:
    from typing import Literal  # Python 3.8

    _unbounded = Literal["unbounded"]
except ImportError:
    _unbounded = str

if "django.contrib.postgres" in settings.INSTALLED_APPS:
    from django.contrib.postgres.fields import ArrayField
else:
    ArrayField = None

__all__ = [
    "ORMPath",
    "XPathMatch",
    "XsdAnyType",
    "XsdAttribute",
    "XsdComplexType",
    "XsdElement",
    "XsdNode",
    "XsdTypes",
    "split_xml_name",
    "FES20",
    "GML21",
    "GML32",
    "XSI",
]

GML21 = "http://www.opengis.net/gml"
GML32 = "http://www.opengis.net/gml/3.2"
XSI = "http://www.w3.org/2001/XMLSchema-instance"
FES20 = "http://www.opengis.net/fes/2.0"

RE_XPATH_ATTR = re.compile(r"\[[^\]]+\]$")  # match [@attr=..]
TYPES_TO_PYTHON = {}


class XsdAnyType:
    """Base class for all types used in the XML definition"""

    name: str
    prefix = None
    is_complex_type = False
    is_geometry = False

    def __str__(self):
        """Return the type name"""
        raise NotImplementedError()

    def with_prefix(self, prefix="xs"):
        xml_name = str(self)
        if ":" in xml_name:
            return xml_name
        else:
            return f"{prefix}:{xml_name}"

    def to_python(self, raw_value):
        """Convert a raw string value to this type representation"""
        return raw_value


class XsdTypes(XsdAnyType, Enum):
    """Brief enumeration of basic XMLSchema types.

    The :class:`XsdElement` and :class:`XsdAttribute` can use these enum members
    to indicate their value is a well-known XML Schema. Some GML types are included as well.

    The default namespace is the "xs:" (XMLSchema).
    Based on https://www.w3.org/TR/xmlschema-2/#built-in-datatypes
    """

    anyType = "anyType"  # Needs to be anyType, as "xsd:any" is an element, not a type.
    string = "string"
    boolean = "boolean"
    decimal = "decimal"  # the base type for all numbers too.
    integer = "integer"  # integer value
    float = "float"
    double = "double"
    time = "time"
    date = "date"
    dateTime = "dateTime"
    anyURI = "anyURI"

    # Number variations
    byte = "byte"  # signed 8-bit integer
    short = "short"  # signed 16-bit integer
    int = "int"  # signed 32-bit integer
    long = "long"  # signed 64-bit integer
    unsignedByte = "unsignedByte"  # unsigned 8-bit integer
    unsignedShort = "unsignedShort"  # unsigned 16-bit integer
    unsignedInt = "unsignedInt"  # unsigned 32-bit integer
    unsignedLong = "unsignedLong"  # unsigned 64-bit integer

    # Less common, but useful nonetheless:
    duration = "duration"
    nonNegativeInteger = "nonNegativeInteger"
    gYear = "gYear"
    hexBinary = "hexBinary"
    base64Binary = "base64Binary"
    token = "token"
    language = "language"

    # Types that contain a GML value as member:
    gmlGeometryPropertyType = "gml:GeometryPropertyType"
    gmlPointPropertyType = "gml:PointPropertyType"
    gmlCurvePropertyType = "gml:CurvePropertyType"  # curve is base for LineString
    gmlSurfacePropertyType = "gml:SurfacePropertyType"  # GML2 had PolygonPropertyType
    gmlMultiSurfacePropertyType = "gml:MultiSurfacePropertyType"
    gmlMultiPointPropertyType = "gml:MultiPointPropertyType"
    gmlMultiCurvePropertyType = "gml:MultiCurvePropertyType"
    gmlMultiGeometryPropertyType = "gml:MultiGeometryPropertyType"

    # Other typical GML values
    gmlCodeType = "gml:CodeType"  # for <gml:name>
    gmlBoundingShapeType = "gml:BoundingShapeType"  # for <gml:boundedBy>

    #: A direct geometry value (used as function argument type)
    gmlAbstractGeometryType = "gml:AbstractGeometryType"

    #: A feature that has an gml:name and gml:boundedBy as possible child element.
    gmlAbstractFeatureType = "gml:AbstractFeatureType"
    gmlAbstractGMLType = "gml:AbstractGMLType"  # base class of gml:AbstractFeatureType

    def __str__(self):
        return self.value

    @property
    def prefix(self) -> str | None:
        """Extrapolate the prefix from the type name"""
        xml_name = str(self)
        colon = xml_name.find(":")
        return xml_name[:colon] if colon else None

    @cached_property
    def is_geometry(self):
        """Whether the value represents a element which contains a GML element."""
        return self.prefix == "gml" and self.value.endswith("PropertyType")

    @cached_property
    def _to_python_func(self):
        if not TYPES_TO_PYTHON:
            _init_types_to_python()

        try:
            return TYPES_TO_PYTHON[self]
        except KeyError:
            raise NotImplementedError(
                f'Casting to "{self}" is not implemented.'
            ) from None

    def to_python(self, raw_value):
        """Convert a raw string value to this type representation"""
        if self.is_geometry:
            # Leave complex values as-is.
            return raw_value

        try:
            return self._to_python_func(raw_value)
        except ExternalParsingError:
            raise  # subclass of ValueError so explicitly caught and reraised
        except (TypeError, ValueError, ArithmeticError) as e:
            # ArithmeticError is base of DecimalException
            raise ExternalParsingError(f"Can't cast '{raw_value}' to {self}.") from e


def _init_types_to_python():
    """Define how well-known scalar types are parsed into python
    (mimicking Django's to_python()):
    """
    global TYPES_TO_PYTHON
    from gisserver.parsers import values  # avoid cyclic import

    as_is = lambda v: v
    TYPES_TO_PYTHON = {
        XsdTypes.date: dateparse.parse_date,
        XsdTypes.dateTime: values.parse_iso_datetime,
        XsdTypes.time: dateparse.parse_time,
        XsdTypes.string: as_is,
        XsdTypes.boolean: values.parse_bool,
        XsdTypes.integer: int,
        XsdTypes.int: int,
        XsdTypes.long: int,
        XsdTypes.short: int,
        XsdTypes.byte: int,
        XsdTypes.unsignedInt: int,
        XsdTypes.unsignedLong: int,
        XsdTypes.unsignedShort: int,
        XsdTypes.unsignedByte: int,
        XsdTypes.float: D,
        XsdTypes.double: D,
        XsdTypes.decimal: D,
        XsdTypes.gmlCodeType: as_is,
        XsdTypes.anyType: values.auto_cast,
    }


class XsdNode:
    """Base class for :class:`XsdElement` and :class:`XsdAttribute`.

    This contains all common mapping/resolving that both elements and attributes share.
    For instance, how XML nodes are mapped into ORM paths, converted into ORM filters,
    parse query input and read model attributes to write as output.
    """

    is_attribute = False
    is_many = False

    name: str
    type: XsdAnyType  # Both XsdComplexType and XsdType are allowed
    prefix: str | None

    #: Which field to read from the model to get the value
    #: This supports dot notation to access related attributes.
    source: models.Field | ForeignObjectRel | None

    #: Which field to read from the model to get the value
    #: This supports dot notation to access related attributes.
    model_attribute: str | None

    def __init__(
        self,
        name: str,
        type: XsdAnyType,
        *,
        prefix: str | None = "app",
        source: models.Field | ForeignObjectRel | None = None,
        model_attribute: str | None = None,
    ):
        # Using plain assignment instead of dataclass turns out to be needed
        # for flexibility and easier subclassing.
        self.name = name
        self.type = type
        self.prefix = prefix
        self.source = source
        self.model_attribute = model_attribute or self.name

        if ":" in self.name:
            raise ValueError("Use 'prefix' argument for namespaces")

        self._attrgetter = operator.attrgetter(self.model_attribute)
        self._valuegetter = self._build_valuegetter(self.model_attribute, self.source)

    def __repr__(self):
        return (
            f"<{self.__class__.__name__}: {self.xml_name}"
            f", model_attribute={self.model_attribute}>"
        )

    @staticmethod
    def _build_valuegetter(
        model_attribute: str,
        field: models.Field | ForeignObjectRel | None,
    ):
        """Select the most efficient read function to retrieves the value."""
        if field is None:
            # No model field, can only use getattr(). The attrgetter function has
            # built-in support for traversing model attributes with dots.
            return operator.attrgetter(model_attribute)
        elif isinstance(field, ForeignObjectRel):
            # Special handling, has no value_from_object()
            return operator.attrgetter(model_attribute)
        elif "." not in model_attribute:
            # Shortcut, can just use Django's value_from_object.
            # This allows Django fields to override the object retrieval.
            # Not using value_to_string() as different output formats may serialize differently.
            return field.value_from_object
        else:
            # Need to traverse foreign key relations before using value_from_objec().
            names = model_attribute.split(".")

            def _related_get_value_from_object(instance):
                for name in names[:-1]:
                    instance = getattr(instance, name)
                return field.value_from_object(instance)

            return _related_get_value_from_object

    @cached_property
    def is_geometry(self) -> bool:
        """Tell whether the XML node/element should be handed as GML geometry."""
        return self.type.is_geometry or isinstance(self.source, GeometryField)

    @cached_property
    def is_array(self) -> bool:
        """Tell whether this node is backed by an PostgreSQL Array Field."""
        return ArrayField is not None and isinstance(self.source, ArrayField)

    @cached_property
    def is_flattened(self) -> bool:
        """Whether the field is a lookup to a relation."""
        return "." in self.model_attribute

    @cached_property
    def xml_name(self):
        """The XML element/attribute name."""
        return f"{self.prefix}:{self.name}" if self.prefix else self.name

    @cached_property
    def orm_path(self) -> str:
        """The ORM field lookup to perform."""
        if self.model_attribute is None:
            raise ValueError(f"Node {self.xml_name} has no 'model_attribute' set.")
        return self.model_attribute.replace(".", "__")

    @cached_property
    def orm_field(self) -> str:
        """The direct ORM field that provides this property."""
        if self.model_attribute is None:
            raise ValueError(f"Node {self.xml_name} has no 'model_attribute' set.")
        return self.model_attribute.split(".", 1)[0]

    @cached_property
    def orm_relation(self) -> tuple[str | None, str]:
        """The ORM field and parent relation"""
        if self.model_attribute is None:
            raise ValueError(f"Node {self.xml_name} has no 'model_attribute' set.")

        try:
            path, field = self.model_attribute.rsplit(".", 1)
        except ValueError:
            return None, self.model_attribute
        else:
            return path.replace(".", "__"), field

    def build_lhs_part(self, compiler: CompiledQuery, match: ORMPath):
        """Give the ORM part when this element is used as left-hand-side of a comparison.
        This is needed for queries like "<element> == <value>"
        """
        return match.orm_path

    def build_rhs_part(self, compiler: CompiledQuery, match: ORMPath):
        """Give the ORM part when this element would be used as right-hand-side.
        This is needed for queries like "<value> == <element>" or "<element> == <element>".
        """
        return F(match.orm_path)

    def get_value(self, instance: models.Model):
        """Provide the value for the data"""
        # For foreign keys, it's not possible to use the model value,
        # as that would conflict with the field type in the XSD schema.
        try:
            if self.type.is_complex_type:
                # This element has sub elements, which need the Django model instance.
                # Avoid unwanted value_from_object(), instead return the model instance.
                value = self._attrgetter(instance)
                if self.is_many and isinstance(value, models.Manager):
                    # Make sure callers can read the individual items by iterating over the value.
                    value = value.all()
                return value
            else:
                # Support value_from_object() on custom fields.
                return self.format_value(self._valuegetter(instance))
        except (AttributeError, ObjectDoesNotExist):
            # E.g. Django foreign keys that point to a non-existing member.
            return None

    def format_value(self, value):
        """Allow to apply some final transformations on a value.
        This is mainly used to support @gml:id which includes a prefix.
        """
        return value

    @cached_property
    def _form_field(self):
        """Internal cached field for to_python()"""
        return self.source.formfield()

    def to_python(self, raw_value: str):
        """Convert a raw value to the Python data type for this element type."""
        try:
            raw_value = self.type.to_python(raw_value)
            if self.source is not None:
                raw_value = self.source.get_prep_value(raw_value)
        except ValidationError as e:
            raise ValidationError(
                f"Invalid data for the '{self.name}' property: {e.messages[0]}",
                code=e.code,
            ) from e
        except (ValueError, TypeError) as e:
            raise ValidationError(
                f"Invalid data for the '{self.name}' property: {e}"
            ) from e

        return raw_value

    def validate_comparison(self, raw_value: str, lookup, tag=None):
        """Validate whether the input value can be used in a comparison.
        This avoids comparing a database DATETIME object to an integer.

        The raw string value can be passed here. Auto-cased values could
        raise an TypeError due to being unsupported by the validation.
        """
        if self.source is not None:
            # Not calling self.source.validate() as that checks for allowed choices,
            # which shouldn't be checked against for a filter query.
            raw_value = self.to_python(raw_value)

            # Check whether the Django model field supports the lookup
            # This prevents calling LIKE on a datetime or float field.
            # For foreign keys, this depends on the target field type.
            if self.source.get_lookup(lookup) is None or (
                isinstance(self.source, RelatedField)
                and self.source.target_field.get_lookup(lookup) is None
            ):
                raise OperationProcessingFailed(
                    "filter",
                    f"Operator '{tag}' is not supported for the '{self.name}' property.",
                    status_code=400,  # not HTTP 500 here. Spec allows both.
                )

        return raw_value


class XsdElement(XsdNode):
    """Declare an XSD element.

    Typically, this maps into a Django model field.

    This holds the definition for a single property in the WFS server.
    It's used in ``DescribeFeatureType`` to output the field metadata,
    and used in ``GetFeature`` to access the actual value from the object.
    Overriding :meth:`get_value` allows to override this logic.

    The :attr:`name` may differ from the underlying :attr:`model_attribute`,
    so the WFS server can use other field names then the underlying model.

    A dotted-path notation can be used for :attr:`model_attribute` to access
    a related field. For the WFS client, the data appears to be flattened.
    """

    nillable: bool | None
    min_occurs: int | None
    max_occurs: int | None

    def __init__(
        self,
        name: str,
        type: XsdAnyType,
        *,
        prefix: str | None = "app",
        nillable: bool | None = None,
        min_occurs: int | None = None,
        max_occurs: int | _unbounded | None = None,
        source: models.Field | ForeignObjectRel | None = None,
        model_attribute: str | None = None,
    ):
        super().__init__(
            name, type, prefix=prefix, source=source, model_attribute=model_attribute
        )
        self.nillable = nillable
        self.min_occurs = min_occurs
        self.max_occurs = max_occurs

    @cached_property
    def as_xml(self):
        attributes = [f'name="{self.name}" type="{self.type}"']
        if self.min_occurs is not None:
            attributes.append(f'minOccurs="{self.min_occurs}"')
        if self.max_occurs is not None:
            attributes.append(f'maxOccurs="{self.max_occurs}"')
        if self.nillable:
            str_bool = "true" if self.nillable else "false"
            attributes.append(f'nillable="{str_bool}"')

        return "<element {} />".format(" ".join(attributes))

    def __str__(self):
        return self.as_xml

    @cached_property
    def is_many(self) -> bool:
        """Tell whether the XML element can be rendered multiple times.
        Note this happens both with array fields and "..._to_many" relations.
        """
        return self.max_occurs and (
            self.max_occurs == "unbounded"
            or self.max_occurs > 1
            # needed for ArrayField(size=1):
            or (ArrayField is not None and isinstance(self.source, ArrayField))
        )


class _XsdElement_WithComplexType(XsdElement):
    """This only exists as "protocol" for the type annotations"""

    type: XsdComplexType


class XsdAttribute(XsdNode):
    """Declare an XSD attribute.

    Typically, this maps into a Django model field.

    Most fields are mapped into XML Elements (:class:`XsdElement`).
    However, WFS also supports XML attributes, and queries against them.
    This class is uses to support filters against attributes like "gml:id".
    """

    is_attribute = True

    type: XsdTypes
    use: str = "optional"

    def __init__(
        self,
        name: str,
        type: XsdAnyType = XsdTypes.string,  # added default
        *,
        prefix: str | None = "app",
        use: str = "optional",
        source: models.Field | ForeignObjectRel | None = None,
        model_attribute: str | None = None,
    ):
        super().__init__(
            name, type, prefix=prefix, source=source, model_attribute=model_attribute
        )
        self.use = use


class GmlIdAttribute(XsdAttribute):
    """A virtual 'gml:id' attribute that can be queried.
    This subclass has overwritten get_value() logic to format the value.
    """

    type_name: str

    def __init__(
        self,
        type_name: str,
        source: models.Field | ForeignObjectRel | None = None,
        model_attribute="pk",
    ):
        super().__init__(
            prefix="gml", name="id", source=source, model_attribute=model_attribute
        )
        object.__setattr__(self, "type_name", type_name)

    def get_value(self, instance: models.Model):
        pk = super().get_value(instance)  # handle dotted-name notations
        return f"{self.type_name}.{pk}"

    def format_value(self, value):
        """Format the value as retrieved from the database."""
        return f"{self.type_name}.{value}"


class GmlNameElement(XsdElement):
    """A subclass to handle the <gml:name> element.
    This displays a human-readable title for the object.

    Currently, this just reads a single attribute,
    but it can be extended to support formatted names
    (although that would make comparisons on ``element@gml:name`` more complex).
    """

    def __init__(
        self,
        model_attribute: str,
        source: models.Field | ForeignObjectRel | None = None,
        feature_type=None,
    ):
        # Prefill most known fields
        super().__init__(
            prefix="gml",
            name="name",
            type=XsdTypes.gmlCodeType,
            min_occurs=0,
            source=source,
            model_attribute=model_attribute,
        )
        self.feature_type = feature_type

    def get_value(self, instance: models.Model):
        """Override value retrieval to retrieve the value from the feature type."""
        if self.feature_type is not None:
            # Let FeatureType provide a nice display/title for the object.
            return self.feature_type.get_display_value(instance)
        else:
            # Fallback, when using this class at a sub-level object.
            return super().get_value(instance)


class GmlBoundedByElement(XsdElement):
    """A subclass to handle the <gml:boundedBy> element.

    This override makes sure this non-model element data
    can be included in the XML tree like every other element.
    Its value is the complete bounding box of the feature type data.
    """

    is_geometry = True  # Override type

    def __init__(self, feature_type):
        # Prefill most known fields
        super().__init__(
            prefix="gml",
            name="boundedBy",
            type=XsdTypes.gmlBoundingShapeType,
            min_occurs=0,
        )
        self.feature_type = feature_type
        self.model_attribute = None

    def build_lhs_part(self, compiler: CompiledQuery, match: ORMPath):
        """Give the ORM part when this element is used as
        left-hand-side of a comparison."""
        return compiler.add_annotation(self.build_rhs_part(compiler, match))

    def build_rhs_part(self, compiler: CompiledQuery, match: ORMPath):
        """Give the ORM part when this element would be used as right-hand-side"""
        raise NotImplementedError("queries against <gml:boundedBy> are not supported")

    def get_value(self, instance: models.Model, crs: CRS | None = None):
        """Provide the value of the <gml:boundedBy> field
        (if this is not given by the database already)."""
        return self.feature_type.get_envelope(instance, crs=crs)


@dataclass(frozen=True)
class XsdComplexType(XsdAnyType):
    """Define an <xsd:complexType> that represents a whole class definition.

    Typically, this maps into a Django model, with each element pointing to a model field.

    The complex can hold multiple :class:`XsdElement` and :class:`XsdAttribute`
    nodes as children, composing an object. The elements themselves can point
    to a complex type themselves, to create a nested class structure.
    That also allows embedding models with their relations into a single response.

    This object definition is the internal "source of truth" regarding
    which field names and field elements are used in the WFS server.
    The ``DescribeFeatureType`` request uses this definition to render the matching XMLSchema.
    Incoming XPath queries are parsed using this object to resolve the XPath to model attributes.

    Objects of this type are typically generated by the ``FeatureType`` and
    ``ComplexFeatureField`` classes, using the Django model data.

    By default, The type is declared as subclass of <gml:AbstractFeatureType>,
    which allows child elements like <gml:name> and <gml:boundedBy>.
    """

    #: Internal class name (without XML prefix)
    name: str

    #: All elements in this class
    elements: list[XsdElement]

    #: All attributes in this class
    attributes: list[XsdAttribute] = field(default_factory=list)

    #: The base class of this type. Typically gml:AbstractFeatureType,
    #: which provides the <gml:name> and <gml:boundedBy> elements.
    base: XsdAnyType = XsdTypes.gmlAbstractFeatureType

    #: The prefix alias to use for the namespace.
    prefix: str = "app"

    #: The Django model class that this type was based on.
    source: type[models.Model] | None = None

    def __str__(self):
        return self.xml_name

    @cached_property
    def xml_name(self):
        """Name in the XMLSchema (e.g. app:SomeClass)."""
        return f"{self.prefix}:{self.name}"

    @property
    def is_complex_type(self):
        return True  # a property to avoid being used as field.

    @cached_property
    def geometry_elements(self) -> list[XsdElement]:
        """Shortcut to get all geometry elements"""
        return [e for e in self.elements if e.is_geometry]

    @cached_property
    def complex_elements(self) -> list[_XsdElement_WithComplexType]:
        """Shortcut to get all elements with a complex type"""
        return [e for e in self.elements if e.type.is_complex_type]

    @cached_property
    def flattened_elements(self) -> list[XsdElement]:
        """Shortcut to get all elements with a flattened model attribite"""
        return [e for e in self.elements if e.is_flattened]

    def resolve_element_path(self, xpath: str) -> list[XsdNode] | None:
        """Resolve a xpath reference to the actual node.
        This returns the whole path, including in-between relations, if a match was found.

        This is used by :meth:`~gisserver.features.FeatureType.resolve_element`
        to convert a request XPath element into the ORM attributes for database queries.
        """
        try:
            pos = xpath.rindex("/")
            node_name = xpath[:pos]
        except ValueError:
            node_name = xpath
            pos = 0

        # Strip any [@attr=..] conditions
        node_name = RE_XPATH_ATTR.sub("", node_name)

        if node_name.startswith("@"):
            # Resolve attributes (e.g. gml:id)
            if pos:
                return None  # invalid attribute

            # Remove app: prefixes, or any alias of it (see explanation below)
            xml_name = node_name[1:]
            attribute = self._find_attribute(xml_name=xml_name)
            return [attribute] if attribute is not None else None
        else:
            element = self._find_element(node_name)
            if element is None:
                return None

            if pos:
                if not element.type.is_complex_type:
                    return None
                else:
                    # Recurse into the child node to find the next part
                    child_path = element.type.resolve_element_path(xpath[pos + 1 :])
                    return [element] + child_path if child_path is not None else None
            else:
                return [element]

    def _find_element(self, xml_name) -> XsdElement | None:
        """Locate an element by name"""
        for element in self.elements:
            if element.xml_name == xml_name:
                return element

        prefix, name = split_xml_name(xml_name)
        if prefix != "gml" and prefix != self.prefix:
            # Ignore current app namespace. Note this should actually compare the
            # xmlns URI's, but this will suffice for now. The ElementTree parser
            # doesn't provide access to 'xmlns' definitions on the element (or it's
            # parents), so a tag like this is essentially not parsable for us:
            # <ValueReference xmlns:tns="http://...">tns:fieldname</ValueReference>
            for element in self.elements:
                if element.name == name:
                    return element

        # When there is a base class, resolve elements there too.
        if self.base.is_complex_type:
            return self.base._find_element(xml_name)
        return None

    def _find_attribute(self, xml_name) -> XsdAttribute | None:
        """Locate an attribute by name"""
        for attribute in self.attributes:
            if attribute.xml_name == xml_name:
                return attribute

        prefix, name = split_xml_name(xml_name)
        if prefix != "gml" and prefix != self.prefix:
            # Allow any namespace to match, since the stdlib ElementTree parser
            # can't resolve namespaces at all.
            for attribute in self.attributes:
                if attribute.name == name:
                    return attribute

        # When there is a base class, resolve attributes there too.
        if self.base.is_complex_type:
            return self.base._find_attribute(xml_name)
        return None


def split_xml_name(xml_name: str) -> tuple[str | None, str]:
    """Remove the namespace prefix from an element."""
    try:
        prefix, name = xml_name.split(":", 1)
        return prefix, name
    except ValueError:
        return None, xml_name


class ORMPath:
    """Base class to provide raw XPath results.

    This base class is designed to allow other query types (besides XPath) too,
    and allows inserting raw data directly to the query compiler (for unit testing).
    """

    def __init__(self, orm_path: str, orm_filters: Q | None = None, is_many=False):
        """Base constructor just assigns items.
        Overwritten classes likely replace this with properties.
        """
        self.orm_path = orm_path
        self.orm_filters = orm_filters
        self.is_many = is_many

    def __repr__(self):
        return (
            f"{self.__class__.__name__}(orm_path={self.orm_path!r}"
            f", orm_filters={self.orm_filters!r})"
        )

    def build_lhs(self, compiler: CompiledQuery):
        """Give the ORM part when this element is used as left-hand-side of a comparison.
        For example: "path == value".
        """
        if self.is_many:
            compiler.add_distinct()
        if self.orm_filters:
            compiler.add_extra_lookup(self.orm_filters)
        return self.orm_path

    def build_rhs(self, compiler: CompiledQuery):
        """Give the ORM part when this element would be used as right-hand-side.
        For example: "path == path" or "value == path".
        """
        if self.is_many:
            compiler.add_distinct()
        return F(self.build_lhs(compiler))


class XPathMatch(ORMPath):
    """The ORM path result from am XPath query.

    This result object defines how to resolve an XPath to a ORM object.
    """

    #: The matched element, with all it's parents.
    nodes: list[XsdNode]

    #: The source XPath query
    query: str

    #: The additional filters are needed (due to [@attr=..] syntax).
    orm_filters: Q | None

    def __init__(self, feature_type: FeatureType, nodes: list[XsdNode], query: str):
        self.feature_type = feature_type
        self.nodes = nodes
        self.query = query
        self.orm_filters = None

        if "[" in self.query:
            # If there is an element[@attr=...]/field tag,
            # the build_...() logic should return a Q() object.
            raise NotImplementedError(
                f"Complex XPath queries are not supported yet: {self.query}"
            )

    @cached_property
    def orm_path(self) -> str:
        """Give the Django ORM path (field__relation__relation2) to the result."""
        return "__".join(xsd_node.orm_path for xsd_node in self.nodes)

    def __iter__(self):
        return iter(self.nodes)

    def __getitem__(self, item) -> XsdNode:
        return self.nodes[item]

    def __repr__(self):
        return f"XPathMatch(nodes={self.nodes!r}, query={self.query!r})"

    @property
    def child(self) -> XsdNode:
        """Return only the final element"""
        return self.nodes[-1]

    @cached_property
    def is_many(self) -> bool:
        """Return whether this ORM path walks over an element that occurs multiple times"""
        return any(node.is_many for node in self.nodes)

    def build_lhs(self, compiler: CompiledQuery):
        """Delegate the LHS construction to the final XsdNode."""
        if self.is_many:
            compiler.add_distinct()
        if self.orm_filters:
            compiler.add_extra_lookup(self.orm_filters)
        return self.child.build_lhs_part(compiler, self)

    def build_rhs(self, compiler: CompiledQuery):
        """Delegate the RHS construction to the final XsdNode."""
        if self.is_many:
            compiler.add_distinct()
        if self.orm_filters:
            compiler.add_extra_lookup(self.orm_filters)
        return self.child.build_rhs_part(compiler, self)


if TYPE_CHECKING:
    from .features import FeatureType
    from .parsers.fes20 import CompiledQuery



from .expressions import ValueReference
from .filters import Filter
from .functions import function_registry
from .identifiers import ResourceId
from .operators import IdOperator
from .query import CompiledQuery
from .sorting import SortBy

__all__ = [
    "Filter",
    "CompiledQuery",
    "ValueReference",
    "ResourceId",
    "IdOperator",
    "SortBy",
    "function_registry",
]


def parse_resource_id_kvp(value) -> IdOperator:
    """Parse the RESOURCEID parameter.
    This returns an IdOperator, as it needs to support multiple pairs.
    """
    return IdOperator([ResourceId(rid) for rid in value.split(",")])



"""Output buffering, to write data in chunks.

This builds on top of StringIO/BytesIO to write data and regularly flush it.
"""
from __future__ import annotations

import io
from typing import Generic, TypeVar

V = TypeVar("V", str, bytes)


class BaseBuffer(Generic[V]):
    """Fast buffer to write data in chunks.
    This avoids performing too many yields in the output writing.
    Especially for GeoJSON, that slows down the response times.
    """

    buffer_class = None

    def __init__(self, chunk_size=8192):
        self.data = self.buffer_class()
        self.chunk_size = chunk_size

    def is_full(self):
        # Calling data.seek() is faster than doing self.size += len(value)
        # because each integer increment is a new object allocation.
        return self.data.tell() >= self.chunk_size

    def write(self, value: V):
        if value is None:
            return
        self.data.write(value)

    def flush(self) -> V:
        """Empty the buffer and return it."""
        data = self.getvalue()
        self.clear()
        return data

    def getvalue(self) -> V:
        return self.data.getvalue()

    def clear(self):
        self.data.seek(0)
        self.data.truncate(0)


class BytesBuffer(BaseBuffer[bytes]):
    """Collect the data as bytes."""

    buffer_class = io.BytesIO

    def __bytes__(self):
        return self.getvalue()


class StringBuffer(BaseBuffer[str]):
    """Collect the data as string"""

    buffer_class = io.StringIO

    def __str__(self):
        return self.getvalue()



"""GML support for the fes filtering.

Overview of GML 3.2 changes: https://mapserver.org/el/development/rfc/ms-rfc-105.html#rfc105
"""

from dataclasses import dataclass
from xml.etree.ElementTree import Element, tostring

from django.contrib.gis.geos import GEOSGeometry

from gisserver.geometries import CRS
from gisserver.parsers.base import tag_registry
from gisserver.parsers.tags import get_attribute
from gisserver.types import GML21, GML32

from .base import AbstractGeometry, TM_Object

_ANY_GML_NS = "{http://www.opengis.net/gml/"


def is_gml_element(element) -> bool:
    """Tell whether the element is an GML element."""
    return element.tag.startswith(_ANY_GML_NS)


@dataclass(repr=False)
@tag_registry.register("Polygon", GML21)
@tag_registry.register("LineString", GML32)
@tag_registry.register("LinearRing", GML32)
@tag_registry.register("MultiLineString", GML32)
@tag_registry.register("MultiPoint", GML32)
@tag_registry.register("MultiPolygon", GML32)
@tag_registry.register("MultiSurface", GML32)
@tag_registry.register("Point", GML32)
@tag_registry.register("Polygon", GML32)
@tag_registry.register("Envelope", GML32)
class GEOSGMLGeometry(AbstractGeometry):
    """Convert the incoming GML into a Django GEOSGeometry"""

    # Not implemented:
    # - Curve
    # - MultiCurve
    # - MultiGeometry
    # - Surface

    xml_ns = ...

    srs: CRS
    geos_data: GEOSGeometry

    @classmethod
    def from_xml(cls, element: Element):
        """Push the whole <gml:...> element into the GEOS parser.
        This avoids having to support the whole GEOS logic.

        GML is a complex beast with many different forms for the same thing:
        http://erouault.blogspot.com/2014/04/gml-madness.html
        """
        srs = CRS.from_string(get_attribute(element, "srsName"))

        # Push the whole <gml:...> element into the GEOS parser.
        # This avoids having to support the whole GEOS logic.
        geos_data = GEOSGeometry.from_gml(tostring(element))
        geos_data.srid = srs.srid
        return cls(srs=srs, geos_data=geos_data)

    def __repr__(self):
        # Better rendering for unit test debugging
        return f"GMLGEOSGeometry(srs={self.srs!r}, geos_data=GEOSGeometry({self.geos_data.wkt!r}))"

    @property
    def wkt(self) -> str:
        """Render the Geometry as well-known text"""
        return self.geos_data.wkt

    @property
    def json(self):
        return self.geos_data.json

    def build_rhs(self, compiler):
        return self.geos_data


@tag_registry.register("After", GML32)
@tag_registry.register("Before", GML32)
@tag_registry.register("Begins", GML32)
@tag_registry.register("BegunBy", GML32)
@tag_registry.register("TContains", GML32)
@tag_registry.register("TEquals", GML32)
@tag_registry.register("TOverlaps", GML32)
@tag_registry.register("During", GML32)
@tag_registry.register("Meets", GML32)
@tag_registry.register("OverlappedBy", GML32)
@tag_registry.register("MetBy", GML32)
@tag_registry.register("EndedBy", GML32)
@tag_registry.register("AnyInteracts", GML32)
class TM_GeometricPrimitive(TM_Object):
    """Not implemented: the whole GML temporal logic"""

    xml_ns = GML32
